@@ -3,6 +3,7 @@
 #include "cmELF.h"
 
 #include <cstddef>
+#include <cstdint>
 #include <map>
 #include <memory>
 #include <sstream>
@@ -10,17 +11,15 @@
 #include <vector>
 
 #include <cm/memory>
+#include <cmext/algorithm>
+
+#include <cm3p/kwiml/abi.h>
 
 #include "cmsys/FStream.hxx"
 
-#include "cm_kwiml.h"
-
-#include "cmAlgorithms.h"
-
 // Include the ELF format information system header.
 #if defined(__OpenBSD__)
 #  include <elf_abi.h>
-#  include <stdint.h>
 #elif defined(__HAIKU__)
 #  include <elf32.h>
 #  include <elf64.h>
@@ -169,6 +168,9 @@ public:
   // Return the recorded ELF type.
   cmELF::FileType GetFileType() const { return this->ELFType; }
 
+  // Return the recorded machine.
+  std::uint16_t GetMachine() const { return this->Machine; }
+
 protected:
   // Data common to all ELF class implementations.
 
@@ -184,6 +186,9 @@ protected:
   // The ELF file type.
   cmELF::FileType ELFType;
 
+  // The ELF architecture.
+  std::uint16_t Machine;
+
   // Whether we need to byte-swap structures read from the stream.
   bool NeedSwap;
 
@@ -380,7 +385,7 @@ private:
 
     // Fix the byte order of the header.
     if (this->NeedSwap) {
-      ByteSwap(x);
+      this->ByteSwap(x);
     }
     return true;
   }
@@ -388,7 +393,7 @@ private:
   {
     if (this->Stream->read(reinterpret_cast<char*>(&x), sizeof(x)) &&
         this->NeedSwap) {
-      ByteSwap(x);
+      this->ByteSwap(x);
     }
     return !this->Stream->fail();
   }
@@ -396,7 +401,7 @@ private:
   {
     if (this->Stream->read(reinterpret_cast<char*>(&x), sizeof(x)) &&
         this->NeedSwap) {
-      ByteSwap(x);
+      this->ByteSwap(x);
     }
     return !this->Stream->fail();
   }
@@ -479,6 +484,8 @@ cmELFInternalImpl<Types>::cmELFInternalImpl(cmELF* external,
     }
   }
 
+  this->Machine = this->ELFHeader.e_machine;
+
   // Load the section headers.
   this->SectionHeaders.resize(this->ELFHeader.e_shnum);
   for (ELF_Half i = 0; i < this->ELFHeader.e_shnum; ++i) {
@@ -574,11 +581,11 @@ std::vector<char> cmELFInternalImpl<Types>::EncodeDynamicEntries(
     dyn.d_un.d_val = static_cast<tagtype>(entry.second);
 
     if (this->NeedSwap) {
-      ByteSwap(dyn);
+      this->ByteSwap(dyn);
     }
 
     char* pdyn = reinterpret_cast<char*>(&dyn);
-    cmAppend(result, pdyn, pdyn + sizeof(ELF_Dyn));
+    cm::append(result, pdyn, pdyn + sizeof(ELF_Dyn));
   }
 
   return result;
@@ -758,6 +765,14 @@ cmELF::FileType cmELF::GetFileType() const
   return FileTypeInvalid;
 }
 
+std::uint16_t cmELF::GetMachine() const
+{
+  if (this->Valid()) {
+    return this->Internal->GetMachine();
+  }
+  return 0;
+}
+
 unsigned int cmELF::GetNumberOfSections() const
 {
   if (this->Valid()) {
