@@ -1,7 +1,6 @@
 /* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
    file Copyright.txt or https://cmake.org/licensing for details.  */
-#ifndef cmMessenger_h
-#define cmMessenger_h
+#pragma once
 
 #include "cmConfigure.h" // IWYU pragma: keep
 
@@ -57,5 +56,3 @@ private:
   bool DevWarningsAsErrors = false;
   bool DeprecatedWarningsAsErrors = false;
 };
-
-#endif
